"""Sparse and TV/HV denoising classes

Classes
-------
SpitfireDenoise

"""

import math
import numpy as np
from .wrappers._spitfire_deconv import (py_spitfire_deconv_2d,
                                        py_spitfire_deconv_3d)


class SpitfireDeconv:
    """Deconvolution of an image using the sparse variation model

    Parameters
    ----------
    regularization: float
        Regularization parameter. It is express in power of 2
        (reg = pow(2, -regularization))
    weighting: float
        weighting parameter between the Hessian and Intensity term
        [0.6, 0.9, 1.0]
    model: str
        Regularization model (SV or HV)
    niter: int
        Maximum number of iterations
    """

    def __init__(self, regularization: float = 12, weighting: float = 0.6,
                 model: str = 'HV', niter: int = 200, deltaz: float = 1.0,
                 deltat: float = 1.0):
        self.regularization = regularization
        self.weighting = weighting
        self.iter = niter
        self.model = model
        self.deltaz = deltaz
        self.deltat = deltat
        self.deconvolved_ = None

    def run(self, image: np.array, psf: np.array):
        im = image.astype(np.float32)
        imin = np.amin(im)
        imax = np.amax(im)
        psf = psf.astype(np.float32)
        psf_ = psf / np.sum(psf)
        im = im / math.sqrt(np.sum(np.square(im)))
        if im.ndim == 2:
            self.deconvolved_ = py_spitfire_deconv_2d(im, psf_,
                                                      self.regularization,
                                                      self.weighting,
                                                      self.model,
                                                      self.iter)
            self.deconvolved_ = self.deconvolved_ * (imax - imin) + imin
        elif im.ndim == 3:
            self.deconvolved_ = py_spitfire_deconv_3d(im, psf_,
                                                      self.regularization,
                                                      self.weighting,
                                                      self.model,
                                                      self.iter,
                                                      self.deltaz)
            self.deconvolved_ = self.deconvolved_ * (imax - imin) + imin
        return self.deconvolved_



"""Helpers to check build environment before build of scikit-bioimaging"""

import os
import sys
import glob
import tempfile
import textwrap
import setuptools  # noqa
import subprocess

from distutils.dist import Distribution
from distutils.sysconfig import customize_compiler
from numpy.distutils.ccompiler import new_compiler
from numpy.distutils.command.config_compiler import config_cc


def _get_compiler():
    """Get a compiler equivalent to the one to build skbioimaging

    Handles compiler specified as follows:
        - python setup.py build_ext --compiler=<compiler>
        - CC=<compiler> python setup.py build_ext
    """
    dist = Distribution({'script_name': os.path.basename(sys.argv[0]),
                         'script_args': sys.argv[1:],
                         'cmdclass': {'config_cc': config_cc}})
    dist.parse_config_files()
    dist.parse_command_line()

    cmd_opts = dist.command_options.get('build_ext')
    if cmd_opts is not None and 'compiler' in cmd_opts:
        compiler = cmd_opts['compiler'][1]
    else:
        compiler = None

    ccompiler = new_compiler(compiler=compiler)
    customize_compiler(ccompiler)

    return ccompiler


def compile_test_program(code, extra_preargs=[], extra_postargs=[]):
    """Check that some C code can be compiled and run"""
    ccompiler = _get_compiler()

    # extra_(pre/post)args can be a callable to make it possible to get its
    # value from the compiler
    if callable(extra_preargs):
        extra_preargs = extra_preargs(ccompiler)
    if callable(extra_postargs):
        extra_postargs = extra_postargs(ccompiler)

    start_dir = os.path.abspath('.')

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            os.chdir(tmp_dir)

            # Write test program
            with open('test_program.c', 'w') as f:
                f.write(code)

            os.mkdir('objects')

            # Compile, test program
            ccompiler.compile(['test_program.c'], output_dir='objects',
                              extra_postargs=extra_postargs)

            # Link test program
            objects = glob.glob(
                os.path.join('objects', '*' + ccompiler.obj_extension))
            ccompiler.link_executable(objects, 'test_program',
                                      extra_preargs=extra_preargs,
                                      extra_postargs=extra_postargs)

            # Run test program
            # will raise a CalledProcessError if return code was non-zero
            output = subprocess.check_output('./test_program')
            output = output.decode(sys.stdout.encoding or 'utf-8').splitlines()
        except Exception:
            raise
        finally:
            os.chdir(start_dir)

    return output


def basic_check_build():
    """Check basic compilation and linking of C code"""
    code = textwrap.dedent(
        """\
        #include <stdio.h>
        int main(void) {
        return 0;
        }
        """)
    compile_test_program(code)



"""Richardson Lucy deconvolution

Classes
-------
RichardsonLucy

"""

import math
import numpy as np
from .wrappers._richardson_lucy_deconv import (py_richardson_lucy_deconv_2d,
                                               py_richardson_lucy_deconv_3d)


class RichardsonLucy:
    """Deconvolution of an image using the Richardson-Lucy algorithm

    Parameters
    ----------
    niter: int
        Maximum number of iterations
    """

    def __init__(self, niter: int = 40):
        self.iter = niter
        self.deconvolved_ = None

    def run(self, image: np.array, psf: np.array):
        im = image.astype(np.float32)
        imin = np.amin(im)
        imax = np.amax(im)
        psf = psf.astype(np.float32)
        psf_ = psf / np.sum(psf)
        im = im / math.sqrt(np.sum(np.square(im)))
        if im.ndim == 2:
            self.deconvolved_ = py_richardson_lucy_deconv_2d(im, psf_,
                                                             self.iter)
            self.deconvolved_ = self.deconvolved_ * (imax - imin) + imin
        elif im.ndim == 3:
            self.deconvolved_ = py_richardson_lucy_deconv_3d(im, psf_,
                                                             self.iter)
            self.deconvolved_ = self.deconvolved_ * (imax - imin) + imin
        return self.deconvolved_



import numpy as np
import math
from .wrappers._psfs import py_gibson_lanni_psf


class PSFGaussian:
    """Generate a Gaussian PSF

    Parameters
    ----------
    sigma: tuple
        Radius of the Gaussian in each dimension
    shape: tuple
        Size of the PSF array in each dimension
    """

    def __init__(self, sigma, shape):
        self.sigma = sigma
        self.shape = shape
        self.psf_ = None

    def run(self):
        """Calculate the PSF image"""
        if len(self.shape) == 2:
            self.psf_ = np.zeros(self.shape)
            x0 = self.shape[0] / 2
            y0 = self.shape[1] / 2
            sigma_x2 = 0.5 / (self.sigma[0] * self.sigma[0])
            sigma_y2 = 0.5 / (self.sigma[1] * self.sigma[1])
            for x in range(self.shape[0]):
                for y in range(self.shape[1]):
                    self.psf_[x, y] = math.exp(- pow(x-x0, 2) * sigma_x2
                                               - pow(y-y0, 2) * sigma_y2)
        elif len(self.shape) == 3:
            x0 = self.shape[0] / 2
            y0 = self.shape[1] / 2
            z0 = self.shape[2] / 2
            sigma_x2 = 0.5 / self.sigma[0] * self.sigma[0]
            sigma_y2 = 0.5 / self.sigma[1] * self.sigma[1]
            sigma_z2 = 0.5 / self.sigma[2] * self.sigma[2]
            for x in range(self.shape[0]):
                for y in range(self.shape[1]):
                    for z in range(self.shape[2]):
                        self.psf_[z, x, y] = math.exp(- pow(x-x0, 2) * sigma_x2
                                                      - pow(y-y0, 2) * sigma_y2
                                                      - pow(z-z0, 2) * sigma_z2)
        else:
            raise Exception('PSFGaussian: can generate only 2D or 3D PSFs')
        return self.psf_


class PSFGibsonLanni:
    """Generate a Gibson-Lanni PSF

    Parameters
    ----------
    shape: tuple
        Size of the PSF array in each dimension
    res_lateral: float
        Lateral resolution
    res_axial: float
        Axial resolution
    numerical_aperture: float
        Numerical aperture
    lambd: float
        Illumination wavelength
    ti0: float
        Working distance
    ni: float
        Refractive index immersion
    ns: float
        Refractive index sample

    """

    def __init__(self, shape, res_lateral=100, res_axial=250,
                 numerical_aperture=1.4, lambd=610,
                 ti0=150, ni=1.5, ns=1.33):
        self.shape = shape
        self.res_lateral = res_lateral
        self.res_axial = res_axial
        self.numerical_aperture = numerical_aperture
        self.lambd = lambd
        self.ti0 = ti0
        self.ni = ni
        self.ns = ns
        self.psf_ = None

    def run(self):
        """Calculate the PSF image"""

        self.psf_ = py_gibson_lanni_psf(self.shape[2], self.shape[1],
                                        self.shape[0],
                                        self.res_lateral, self.res_axial,
                                        self.numerical_aperture, self.lambd,
                                        self.ti0, self.ni, self.ns)
        self.psf_ = np.transpose(self.psf_, (2, 0, 1))
        return self.psf_



import sys
import os

from sdeconv._build_utils import cythonize_extensions


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    import numpy

    libraries = []
    if os.name == 'posix':
        libraries.append('m')

    config = Configuration('sdeconv', parent_package, top_path)

    # submodules with build utilities
    config.add_subpackage('__check_build')
    config.add_subpackage('_build_utils')

    # submodules which have their own setup.py
    config.add_subpackage('deconv')

    # Skip cythonization as we do not want to include the generated
    # C/C++ files in the release tarballs as they are not necessarily
    # forward compatible with future versions of Python for instance.
    if 'sdist' not in sys.argv:
        cythonize_extensions(top_path, config)

    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



from .spitfire import SpitfireDeconv
from .richardson_lucy import RichardsonLucy
from .wiener import WienerDeconv
from .psfs import PSFGaussian, PSFGibsonLanni

__all__ = ['SpitfireDeconv',
           'RichardsonLucy',
           'WienerDeconv',
           'PSFGaussian',
           'PSFGibsonLanni']



""" Distributor init file

Distributors: you can add custom code here to support particular distributions
of sdeconv.

For example, this is a good place to put any checks for hardware requirements.

The sdeconv standard source distribution will not put code in this
file, so you can safely replace this file with your own version.
"""
